use rand::prelude::SliceRandom;
use rand::Rng;
use petgraph::graph::UnGraph;
use petgraph::graph::Graph;
use petgraph::data::*;
use petgraph::algo::min_spanning_tree;
use rand::rngs::ThreadRng;

use crate::level::{Level, LevelElement, RoomType};

const ROOMS_VERTICAL: usize = 7;
const ROOMS_HORIZONTAL: usize = 4;
const ROOM_WIDTH: usize = 7;
const ROOM_HEIGHT: usize = 6;

#[derive(Copy, Clone, Debug, PartialEq)]
struct Room {
    pub kind: RoomType,
    pub offset_x: usize,
    pub offset_y: usize,
    pub width: usize,
    pub height: usize,
    // pub connection_down: bool,
    // pub connection_right: bool,
    // pub connected: bool,
}

impl Room {
    fn new() -> Self {
        Self {
            kind: RoomType::EmptyRoom,
            offset_x: 0,
            offset_y: 0,
            width: 0,
            height: 0,
            // connection_down: false,
            // connection_right: false,
            // connected: false,
        }
    }
    /// change the size and position of a room randomly within its bounds
    fn random(&mut self, rng: &mut ThreadRng) {
        let width = rng.gen_range(3..6);
        let height = rng.gen_range(3..5);
        self.width = width;
        self.height = height;
        self.offset_x = rng.gen_range(0..(ROOM_WIDTH - width));
        self.offset_y = rng.gen_range(0..(ROOM_HEIGHT - height));
    }
}

#[derive(Debug)]
pub struct LevelGenerator {
    rooms: [[Room; ROOMS_HORIZONTAL]; ROOMS_VERTICAL],
}

impl LevelGenerator {
    pub fn generate(level: usize) -> Self {
        let mut rng = rand::thread_rng();
        let mut rooms = [[Room::new(); ROOMS_HORIZONTAL]; ROOMS_VERTICAL];
        let mut graph = UnGraph::<(usize, usize), u16>::default();

        // trick the room_connectable function into failing on the first iteration
        rooms[0][0].kind = RoomType::BasicRoom;

        while !LevelGenerator::rooms_connectable(&rooms) {
            let mut room_types: Vec<RoomType> = Vec::with_capacity(ROOMS_HORIZONTAL * ROOMS_VERTICAL);
            // level 0 contains a start room, all others contain a stair up
            if level == 0 {
                room_types.push(RoomType::Start);
            } else {
                room_types.push(RoomType::StairUp);
            }
            // level 24 (the last) contains an end room, all others a stair down
            if level == 24 {
                room_types.push(RoomType::End);
            } else {
                room_types.push(RoomType::StairDown);
            }
            // generate a random set of rooms and shuffle them
            for _ in room_types.len()..ROOMS_HORIZONTAL * ROOMS_VERTICAL {
                match rng.gen_range(1..=6) {
                    // TODO tune room type distribution
                    1..=3 => { room_types.push(RoomType::EmptyRoom) }
                    _ => { room_types.push(RoomType::BasicRoom) }
                }
            }
            room_types.shuffle(&mut rng);

            graph.clear();
            // place the rooms in the array an add nodes to the graph for every non empty room
            for c in 0..ROOMS_VERTICAL {
                for r in 0..ROOMS_HORIZONTAL {
                    rooms[c][r].kind = room_types.pop().unwrap();
                    if rooms[c][r].kind != RoomType::EmptyRoom {
                        rooms[c][r].random(&mut rng);
                        graph.add_node((c, r));
                    }
                }
            }
        }

        // add edges to the graph connecting each room to all of its neighbours (max 4 of them)
        for c in 0..ROOMS_VERTICAL {
            for r in 0..ROOMS_HORIZONTAL {
                if rooms[c][r].kind == RoomType::EmptyRoom {
                    continue;
                }
                let src_index = graph.node_indices().find(|i| graph[*i] == (c, r)).unwrap();
                for r_1 in r + 1..ROOMS_HORIZONTAL {
                    if rooms[c][r_1].kind != RoomType::EmptyRoom {
                        let tgt_index = graph.node_indices().find(|i| graph[*i] == (c, r_1)).unwrap();
                        // todo use random weight for edge
                        graph.add_edge(src_index, tgt_index, 1);
                        break;
                    }
                }
                for c_1 in c + 1..ROOMS_VERTICAL {
                    if rooms[c_1][r].kind != RoomType::EmptyRoom {
                        let tgt_index = graph.node_indices().find(|i| graph[*i] == (c_1, r)).unwrap();
                        // todo use random weight for edge
                        graph.add_edge(src_index, tgt_index, 1);
                        break;
                    }
                }
            }
        }

        // calculate a minimum spanning tree
        let mst: Graph<(usize, usize), u16, petgraph::Undirected> = Graph::from_elements(min_spanning_tree(&graph));
        for edge in mst.raw_edges() {
            let src = mst[edge.source()];
            let tgt = mst[edge.target()];
            println!("{:?} {:?}", src, tgt);
        }
        // println!("Level {}\n{}", level, Dot::new(&mst));

        LevelGenerator {
            rooms
        }
    }
    /// Verifies that for a given matrix of rooms each room has at least one other room in the
    /// same row or column.
    fn rooms_connectable(rooms: &[[Room; ROOMS_HORIZONTAL]; ROOMS_VERTICAL]) -> bool {
        for c in 0..ROOMS_VERTICAL {
            for r in 0..ROOMS_HORIZONTAL {
                if rooms[c][r].kind != RoomType::EmptyRoom {
                    let mut connected = 0;
                    for c1 in 0..ROOMS_VERTICAL {
                        if rooms[c1][r].kind != RoomType::EmptyRoom {
                            connected += 1;
                        }
                    }
                    for r1 in 0..ROOMS_HORIZONTAL {
                        if rooms[c][r1].kind != RoomType::EmptyRoom {
                            connected += 1;
                        }
                    }
                    if connected <= 2 {
                        return false;
                    }
                }
            }
        }
        return true;
    }
    pub fn render(&self) -> Level {
        let mut rng = rand::thread_rng();
        let mut s = [[LevelElement::Wall; 1 + ROOMS_HORIZONTAL * ROOM_HEIGHT]; 1 + ROOMS_VERTICAL * ROOM_WIDTH];
        let mut start_x: usize = 0;
        let mut start_y: usize = 0;
        for c in 0..ROOMS_VERTICAL {
            for r in 0..ROOMS_HORIZONTAL {
                let top = 1 + r * ROOM_HEIGHT;
                let left = 1 + c * ROOM_WIDTH;
                let room = self.rooms[c][r];
                for x in 0..room.width {
                    for y in 0..room.height {
                        s[left + room.offset_x + x][top + room.offset_y + y] = LevelElement::Floor;
                    }
                }
                if room.kind == RoomType::StairDown {
                    s[left + room.offset_x + rng.gen_range(0..room.width)][top + room.offset_y + rng.gen_range(0..room.height)] = LevelElement::StairDown;
                }
                if room.kind == RoomType::Start || room.kind == RoomType::StairUp {
                    start_x = left + room.offset_x + rng.gen_range(0..room.width);
                    start_y = top + room.offset_y + rng.gen_range(0..room.height);
                }
                if room.kind == RoomType::StairUp {
                    s[start_x][start_y] = LevelElement::StairUp;
                }
            }
        }
        Level {
            structure: s,
            start_x,
            start_y,
        }
    }
}

#[test]
fn test_level_gen() {
    let level = LevelGenerator::generate(0);
    println!("{:?}", level);
    assert_eq!(1, 2);
}